# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# URL:    https://gong.host
# Date:   20230417

from neetbox.pipeline._signal_and_slot import listen, watch

__all__ = ["watch", "listen"]
